/*
  Copyrights 2005-2011 Sisulizer

  A class that translates forms of a Windows Forms application.
*/

// If you use .NET 1.x remove comments on the next line
//#define NET1

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Globalization;
using System.Threading;
using System.Text;
using System.Windows.Forms;

namespace Sisulizer.WindowsForms
{
  public class Translator
  {
    static private bool listViewItemsTranslationEnabled = true;
    ComponentResourceManager resources = null;
    System.Windows.Forms.ToolTip toolTip = null;

    public Translator()
    {
      FormCollection forms = Application.OpenForms;

      for (int i = 0; i < forms.Count; i++)
      {
        TranslateForm(forms[i]);
      }
    }

    public Translator(Form form)
    {
      TranslateForm(form);
    }

    private void TranslateGridViewColumns(DataGridViewColumnCollection items)
    {
      for (int i = 0; i < items.Count; i++)
      {
        DataGridViewColumn item = items[i];
        resources.ApplyResources(item, item.Name);
      }
    }

    private string GetName(string name, object tag)
    {
      if ((name == "") && (tag is string))
        name = (string)tag;

      return name;
    }

    private void TranslateListViewColumns(ListView.ColumnHeaderCollection items)
    {
      for (int i = 0; i < items.Count; i++)
      {
        ColumnHeader item = items[i];
        resources.ApplyResources(item, GetName(item.Name, item.Tag));
      }
    }

    private void TranslateListViewItems(ListView.ListViewItemCollection items)
    {
      if (!ListViewItemsTranslationEnabled)
        return;

      for (int i = 0; i < items.Count; i++)
      {
        ListViewItem item = items[i];
        string name = item.ListView.Name + ".Items";

        if (i > 0)
          name = name + i.ToString();

        ListViewItem source = resources.GetObject(name) as ListViewItem;
        item.Text = source.Text;
        item.ToolTipText = source.ToolTipText;

        for (int j = 0; j < item.SubItems.Count; j++)
          item.SubItems[j].Text = source.SubItems[j].Text;
      }
    }

#if !NET1
    private void TranslateToolStripItems(ToolStripItemCollection items)
    {
      for (int i = 0; i < items.Count; i++)
      {
        ToolStripItem item = items[i];
        resources.ApplyResources(item, item.Name);

        if (item is ToolStripDropDownItem)
          TranslateToolStripItems((item as ToolStripDropDownItem).DropDownItems);
      }
    }
#endif

    private void TranslateControl(Control control, string name)
    {
      // Translate child controls
      if (control.Controls.Count > 0)
      {
        for (int i = 0; i < control.Controls.Count; i++)
        {
          Control childControl = control.Controls[i];
          TranslateControl(childControl, childControl.Name);
        }
      }

      // Translate this control
      resources.ApplyResources(control, name);
      toolTip.SetToolTip(control, resources.GetString(name + ".ToolTip"));

      // Translate parts of control that does not automatically translate when using ApplyResources
      if (control is DataGridView)
        TranslateGridViewColumns((control as DataGridView).Columns);
      else if (control is ListView)
      {
        TranslateListViewColumns((control as ListView).Columns);
        TranslateListViewItems((control as ListView).Items);
      }
#if !NET1
      else if (control is MenuStrip)
        TranslateToolStripItems((control as MenuStrip).Items);
#endif
    }

    private void LayoutControl(Control control)
    {
      control.ResumeLayout(false);
      control.PerformLayout();
    }

    private void TranslateForm(Form form)
    {
      int left = form.Left;
      int top = form.Top;
      int width = form.Width;
      int height = form.Height;

      toolTip = new System.Windows.Forms.ToolTip();
      resources = new ComponentResourceManager(form.GetType());

      // Translate controls
      TranslateControl(form, "$this");

      form.Left = left;
      form.Top = top;
      form.Width = width;
      form.Height = height;

      // Layout controls
      for (int i = 0; i < form.Controls.Count; i++)
        LayoutControl(form.Controls[i]);
    
      LayoutControl(form);
    }

    static public bool ListViewItemsTranslationEnabled
    {
      get { return listViewItemsTranslationEnabled; }
      set { listViewItemsTranslationEnabled = value; }
    }

    // Translate all forms
    static public void Translate()
    {
      new Translator();
    }

    // Translate only the given form
    static public void Translate(Form form)
    {
      new Translator(form);
    }

    // Set a new UI language and translate all existing forms.
    static public void SetUILanguage(CultureInfo language)
    {
      Thread.CurrentThread.CurrentUICulture = language;
      Translator.Translate();
    }

    // Mirrors the child controls of a control
    static public void FlipControl(Control control, bool includeChildren)
    {
      for (int i = 0; i < control.Controls.Count; i++)
      {
        Control thisControl = control.Controls[i];
        thisControl.Left = control.Width - (thisControl.Left + thisControl.Width);

        if (includeChildren)
          FlipControl(thisControl, includeChildren);
      }
    }
  }
}