{
  Copyrights 2005-2011 Sisulizer

  A dialog that shows the available resource DLLs and lets the user to select one.
  This compiles with Delphi 5 and later.
}

{$I LaDef.inc}

unit LaDialog;

interface

uses
  Classes, Controls, Dialogs, Forms, StdCtrls, SysUtils, Windows, LaResource, LaCommon;

type
  TLaLanguageDialog = class(TForm)
    List: TListBox;
    OkButton: TButton;
    CancelButton: TButton;
    procedure FormCreate(sender: TObject);
    procedure FormShow(sender: TObject);
    procedure FormDestroy(sender: TObject);
    procedure ListDrawItem(control: TWinControl; index: Integer; rect: TRect; state: TOwnerDrawState);
    procedure ListDblClick(sender: TObject);

  private
    FCompatible: Boolean;
    FOriginalLanguage: String;
    FUseOriginalLanguage: Boolean;

    function GetLanguage(i: Integer): String;
    function GetLanguageCount: Integer;
    function GetSelectedLangId: Integer;
    function GetSelectedLanguage: String;

    procedure SetSelectedLanguage(const value: String);

    procedure AddString(const name: String; const code: String);
{$IFNDEF UNICODE_VCL}
    procedure AddUnicodeString(const name: UnicodeString; const code: AnsiString);
{$ENDIF}

  public
    property LanguageCount: Integer read GetLanguageCount;
    property Languages[i: Integer]: String read GetLanguage;
    property Compatible: Boolean read FCompatible write FCompatible;
    property OriginalLanguage: String read FOriginalLanguage write FOriginalLanguage;
    property UseOriginalLanguage: Boolean read FUseOriginalLanguage write FUseOriginalLanguage;
    property SelectedLanguage: String read GetSelectedLanguage write SetSelectedLanguage;
    property SelectedLangId: Integer read GetSelectedLangId;
  end;

  // Specifies the options of the select language dialog
  TLaDialogOption =
  (
    doUseCurrentLanguage,  // If included show language names in the current language. If excluded show language names in native language.
    doCompatible,          // Show only those languages that are compatible to the current system settings
    doCheckResourceCount   // If included check that reosurce DLLs exists
  );
  TLaDialogOptions = set of TLaDialogOption;

// Shows a dialog that contains a list of available resource DLLs. Lets the user
// to select a new language and finally turns that language on.
// - originalLanguage  specifies the Windows language code of the original language of the applications.
// - dialogOptions     specifies the select language dialog options. See TLaDialogOption
// - resourceOptions   specifies the select language dialog options. See Laresource.TLaResourceOption
// - position          specifies the initial position of the select language dialog
function SelectResourceLocale(
  const originalLanguage: String = '';
  dialogOptions: TLaDialogOptions = [];
  resourceOptions: TLaResourceOptions = [];
  position: TPosition = poMainFormCenter): Boolean; overload;

// Here NSL language ids are used instead of Windows language code
function SelectResourceLocale(
  originalLanguage: Integer = 0;
  dialogOptions: TLaDialogOptions = [];
  resourceOptions: TLaResourceOptions = [];
  position: TPosition = poScreenCenter): Boolean; overload;

// As above but does not turn on the selected language. Just shows the language
// dialog and asks the user to select a new language.
// - language returns the selected language using either Windows language code or
//   NLS langiage id
//
// Other parameters are as in SelectResourceLocale
function AskResourceLocale(
  var language: String;
  originalLanguage: String = '';
  options: TLaDialogOptions = [];
  position: TPosition = poScreenCenter): Boolean; overload;

function AskResourceLocale(
  var language: Integer;
  originalLanguage: Integer = 0;
  options: TLaDialogOptions = [];
  position: TPosition = poScreenCenter): Boolean; overload;

// Checks if the language of currently loaded resource file is compatible to the
// system code page of the computer. If not shows an error message.
procedure CheckCodePage;

implementation

{$R *.dfm}

uses
  LaTranslator;

function SelectResourceLocale(
  const originalLanguage: String;
  dialogOptions: TLaDialogOptions;
  resourceOptions: TLaResourceOptions;
  position: TPosition): Boolean;
var
  language: String;
begin
  Result := AskResourceLocale(language, originalLanguage, dialogOptions, position);

  if Result then
  begin
    if language = originalLanguage then
      language := '';

    Result := SetNewResourceFile(language, resourceOptions, originalLanguage);
  end;
end;

function SelectResourceLocale(
  originalLanguage: Integer;
  dialogOptions: TLaDialogOptions;
  resourceOptions: TLaResourceOptions;
  position: TPosition): Boolean;
var
  language: Integer;
begin
  Result := AskResourceLocale(language, originalLanguage, dialogOptions, position);

  if Result then
  begin
    if Integer(language) = originalLanguage then
      language := 0;

    Result := SetNewResourceFile(language, resourceOptions, originalLanguage);
  end;
end;

function AskResourceLocale(
  var language: String;
  originalLanguage: String;
  options: TLaDialogOptions;
  position: TPosition): Boolean;
var
  dialog: TLaLanguageDialog;
begin
  if doCheckResourceCount in options then
    CheckResourceCount;

  dialog := TLaLanguageDialog.Create(nil);
  try
    dialog.Position := position;
    dialog.UseOriginalLanguage := not (doUseCurrentLanguage in options);
    dialog.Compatible := doCompatible in options;
    dialog.OriginalLanguage := originalLanguage;

    Result := dialog.ShowModal = mrOk;

    if Result then
      language := dialog.SelectedLanguage;
  finally
    dialog.Free;
  end;
end;

function AskResourceLocale(
  var language: Integer;
  originalLanguage: Integer;
  options: TLaDialogOptions;
  position: TPosition): Boolean;
var
  dialog: TLaLanguageDialog;
begin
  dialog := TLaLanguageDialog.Create(nil);
  try
    dialog.Position := position;
    dialog.UseOriginalLanguage := not (doUseCurrentLanguage in options);
    dialog.Compatible := doCompatible in options;
    dialog.OriginalLanguage := LocaleToExtension(originalLanguage);

    Result := dialog.ShowModal = mrOk;

    if Result then
      language := dialog.SelectedLangId;
  finally
    dialog.Free;
  end;
end;

procedure CheckCodePage;
{$IFNDEF UNICODE_VCL}
resourcestring
  SIncompatibleCodePage =
    'The current system code page %0:d is not compatible to %1:s.'#13#10 +
    'Text on the form will render incorrectly.';
var
  langId: Integer;
{$ENDIF}
begin
{$IFNDEF UNICODE_VCL}
  langId := GetActiveLangId(LANG_ENGLISH);

  if (langId <> LANG_ENGLISH) and not IsLocaleCompatible(langId) then
    ShowMessage(Format(
      SIncompatibleCodePage,
      [GetACP, GetWindowsLocaleName(GetActiveLangId(LANG_ENGLISH), False)]));
{$ENDIF}
end;


// TLaLanguageDialog

function TLaLanguageDialog.GetLanguageCount: Integer;
begin
  Result := List.Items.Count;
end;

function TLaLanguageDialog.GetLanguage(i: Integer): String;
begin
  Result := TLanguageListItem(List.Items.Objects[i]).Code;
end;

function TLaLanguageDialog.GetSelectedLangId: Integer;
begin
  Result := ExtensionToLocale(SelectedLanguage);
end;

function TLaLanguageDialog.GetSelectedLanguage: String;
begin
  if List.ItemIndex >= 0 then
    Result := Languages[List.ItemIndex]
  else
    Result := '';
end;

procedure TLaLanguageDialog.SetSelectedLanguage(const value: String);
var
  i: Integer;
begin
  for i := 0 to LanguageCount - 1 do
    if Languages[i] = value then
    begin
      List.ItemIndex := i;
      Break;
    end;
end;

procedure TLaLanguageDialog.AddString(const name: String; const code: String);
begin
  List.Items.AddObject(name, TLanguageListItem.Create(code, name));
end;

{$IFNDEF UNICODE_VCL}
procedure TLaLanguageDialog.AddUnicodeString(const name: UnicodeString; const code: AnsiString);
begin
  List.Items.AddObject(name, TLanguageListItem.Create(code, name));
end;
{$ENDIF}

procedure TLaLanguageDialog.FormCreate(Sender: TObject);
begin
  FCompatible := False;
  FOriginalLanguage := '';
  FUseOriginalLanguage := False;

  if roFlipChildren in LaResourceOptions then
    InitializeForm(Self);
end;

procedure TLaLanguageDialog.FormShow(Sender: TObject);

  procedure Process(const code: String);
  var
    locale: Integer;
  begin
    locale := ExtensionToLocale(code);

{$IFNDEF UNICODE_VCL}
    if IsNtOrLater then
      AddUnicodeString(GetWindowsLocaleWideName(locale, FUseOriginalLanguage), code)
    else
{$ENDIF}
      AddString(GetWindowsLocaleName(locale, FUseOriginalLanguage), code);
  end;

var
  i: Integer;
  thisList: TStringList;
begin
  thisList := TStringList.Create;
  try
    GetAvailableResources(thisList, '', FCompatible);

    if OriginalLanguage <> '' then
      Process(OriginalLanguage);

    for i := 0 to thisList.Count - 1 do
      Process(thisList[i]);
  finally
    thisList.Free;
  end;

  for i := 0 to LanguageCount - 1 do
  begin
    if ((LoadedResourceLocale <> '') and (CompareText(Languages[i], LoadedResourceLocale) = 0)) or
      ((LoadedResourceLocale = '') and (CompareText(Languages[i], OriginalLanguage) = 0)) then
    begin
      List.ItemIndex := i;
      Break;
    end;
  end;

  if (List.ItemIndex = -1) and (List.Items.Count > 0) then
    List.ItemIndex := 0;
end;

procedure TLaLanguageDialog.FormDestroy(Sender: TObject);
var
  i: Integer;
begin
  for i := 0 to List.Items.Count - 1 do
    TLanguageListItem(List.Items.Objects[i]).Free;
end;

procedure TLaLanguageDialog.ListDrawItem(
  control: TWinControl;
  index: Integer;
  rect: TRect;
  state: TOwnerDrawState);
begin
  TLanguageListItem.DrawItem(Self, List, index, rect, state);
end;

procedure TLaLanguageDialog.ListDblClick(Sender: TObject);
begin
  ModalResult := mrOk;
end;

end.
