{
  Copyright 2005-2011 Sisulizer

  Unit that translates forms.
  This compiles with Delphi 5 and later.
}

{$I LaDef.inc}

unit LaTranslator;

interface

uses
  Classes,
  Forms,
  TypInfo,
  LaCommon,
  LaResource;

const
  STRING_TYPES_C = [tkString, tkLString, tkWString{$IFDEF UNICODE_VCL}, tkUString{$ENDIF}];

type
  TLaTranslator = class(TObject)
  private
    FCurrent: TComponent;
    FHost: TComponent;
    FIndex: Integer;
    FName: String;
    FObj: TObject;
    FPropertyName: String;
    FPropInfo: PPropInfo;
    FReader: TReader;
    FSubName: String;
    FUnnamedTypes: TStringList;
    FTranslateLayout: Boolean;

    function GetTypeInfo: PTypeInfo;

    function ProcessBinary: RawByteString;
    procedure ProcessBytes(count: Integer);
    procedure ProcessCollection;
    procedure ProcessComponent(parent: TComponent; root: Boolean);
    procedure ProcessList;
    procedure ProcessProperty(obj: TObject);
    procedure ProcessSet;
    procedure ProcessPropertyValue;

    function GetPropValue(instance: TObject): Variant;
    procedure SetPropValue(instance: TObject; const value: Variant);

    procedure DoTranslate(component: TComponent; resourceName: String = '');

  protected
    procedure Translate(component: TComponent);
    procedure TranslateForm(form: TForm);

    property TypeInfo: PTypeInfo read GetTypeInfo;

  public
    constructor Create;
    destructor Destroy; override;

    class procedure TranslateForms;
  end;

  TLaStream = class(TMemoryStream)
  public
    constructor Create(const value: RawByteString);

    function ReadByte: Byte;
    function ReadWord: Word;
    function ReadInteger: Integer;
    function ReadPointer: Pointer;
    function ReadShortString: RawByteString;
    function ReadShortUnicodeString: UnicodeString;
  end;

  TLaModule = class(TObject)
  protected
    function IgnoreProperty(
      component: TComponent;
      obj: TObject;
      current: Variant;
      var value: Variant;
      propInfo: PPropInfo = nil): Boolean;

    procedure AfterTranslate(
      component: TComponent;
      obj: TObject;
      propInfo: PPropInfo = nil);

  public
    function CanTranslate(obj: TObject): Boolean; virtual; abstract;

    procedure Translate(
      component: TComponent;
      obj: TObject;
      const name: String;
      value: Variant;
      index: Integer); virtual; abstract;

    function GetActualObject(obj: TObject; const propName: String): TObject; virtual;
    function GetActualName(obj: TObject; const propName: String): String; virtual;
  end;

  TLaModuleClass = class of TLaModule;

  TLaStringsModule = class(TLaModule)
  public
    function CanTranslate(obj: TObject): Boolean; override;

    procedure Translate(
      component: TComponent;
      obj: TObject;
      const name: String;
      value: Variant;
      index: Integer); override;
  end;

  TLaModules = class(TObject)
  private
    FItems: TList;

    function GetCount: Integer;
    function GetItem(i: Integer): TLaModule;

    procedure ClearItems;

  public
    constructor Create;
    destructor Destroy; override;

    procedure Register(moduleClass: TLaModuleClass);
    function CanTranslate(obj: TObject; var module: TLaModule): Boolean;

    function GetActualObject(obj: TObject; const propName: String): TObject;
    function GetActualName(obj: TObject; const propName: String): String;

    property Count: Integer read GetCount;
    property Items[i: Integer]: TLaModule read GetItem; default;
  end;

  // This event is called before translating a property.
  // host         is either form or data module where the object belongs to.
  // obj          is the component or part of component that is currenty translated.
  // propertyInfo contains the property information such as name and type.
  // currentValue contains the current value of the property.
  // newValue     contains the translated value of the property. You can change it.
  // cancel       is False by default. Set it True if you want to cancel the translation.
  TLaBeforeTranslateEvent = procedure(
    host: TComponent;
    obj: TObject;
    propertyInfo: PPropInfo;
    const currentValue: Variant;
    var newValue: Variant;
    var cancel: Boolean);

  // This event is called after translating a property
  // See information about parameters above.
  TLaAfterTranslateEvent = procedure(
    item: TComponent;
    obj: TObject;
    propertyInfo: PPropInfo);

var
  LaEnabledProperties: TTypeKinds;
  LaFormPositionTranslationEnabled: Boolean;
  LaBeforeTranslate: TLaBeforeTranslateEvent;
  LaAfterTranslate: TLaAfterTranslateEvent;
  LaModules: TLaModules;

implementation

uses
  Controls,
  SysConst,
  SysUtils,
  Graphics,
  ActnList,
  Menus,
{$IFDEF VCL5}
  ActiveX,
  StdCtrls,
{$ENDIF}
{$IFDEF VCL6}
  Variants,
{$ENDIF}
  Windows;


// TLaTranslator

constructor TLaTranslator.Create;
begin
  inherited;
  FUnnamedTypes := TStringList.Create;
end;

destructor TLaTranslator.Destroy;
begin
  FUnnamedTypes.Free;
  inherited;
end;

function TLaTranslator.GetTypeInfo: PTypeInfo;
begin
  if (FPropInfo <> nil) and (FPropInfo.PropType <> nil) then
    Result := FPropInfo.PropType^
  else
    Result := nil;
end;

function TLaTranslator.ProcessBinary: RawByteString;
var
  size: Integer;
begin
  FReader.ReadValue;
  FReader.Read(size, SizeOf(size));

  if size > 0 then
  begin
    SetLength(Result, size);
    FReader.Read(PAnsiChar(Result)^, Length(Result));
  end
  else
    Result := '';
end;

procedure TLaTranslator.ProcessBytes(count: Integer);
var
  data: AnsiString;
begin
  FReader.ReadValue;

  if count > 0 then
  begin
    SetLength(data, count);
    FReader.Read(PAnsiChar(data)^, count);
  end;
end;

procedure TLaTranslator.ProcessCollection;
var
  i: Integer;
  thisObj, actualObj: TObject;
  propInfo: PPropInfo;
  item: TCollectionItem;
  collection: TCollection;
begin
  if (TypeInfo <> nil) and (TypeInfo.Kind <> tkClass) then
    raise Exception.Create('Invalid data');

  actualObj := FObj;

  if actualObj <> nil then
  begin
    propInfo := GetPropInfo(actualObj, FName);

    if (propInfo = nil) and (FSubName <> '') then
      propInfo := GetPropInfo(actualObj, FSubName);

    if (propInfo = nil) then
    begin
      thisObj := LaModules.GetActualObject(actualObj, FName);

      if thisObj <> nil then
      begin
        actualObj := thisObj;
        propInfo := GetPropInfo(actualObj, FName);
      end;
    end;
  end
  else
    propInfo := nil;

  if propInfo <> nil then
    collection := TObject(GetOrdProp(actualObj, propInfo)) as TCollection
  else
    collection := nil;

  FReader.ReadValue;
  i := 0;

  while not FReader.EndOfList do
  begin
    if (collection <> nil) and (i < collection.Count) then
      item := collection.Items[i]
    else
      item := nil;

    if FReader.NextValue in [vaInt8, vaInt16, vaInt32] then
      FReader.ReadInteger;

    FReader.ReadListBegin;

    while not FReader.EndOfList do
      ProcessProperty(item);

    FReader.ReadListEnd;

    Inc(i);
  end;

  FReader.ReadListEnd;
end;

procedure TLaTranslator.ProcessList;
begin
  FReader.ReadListBegin;
  FIndex := 0;

  while not FReader.EndOfList do
  begin
    ProcessPropertyValue;
    Inc(FIndex);
  end;

  FReader.ReadListEnd;
end;

procedure TLaTranslator.ProcessSet;
var
  ident: String;
begin
  FReader.ReadValue;

  while True do
  begin
    ident := FReader.ReadStr;

    if ident = '' then
      Break;
  end;
end;

procedure TLaTranslator.ProcessProperty(obj: TObject);
var
  p: Integer;
  propName, actualName: String;
  objNames: TStringList;
  i: Integer;
begin
  FPropertyName := FReader.ReadStr;
  FName := FPropertyName;
  FSubName := '';
  FObj := obj;

  if FObj <> nil then
  begin
    FPropInfo := GetPropInfo(FObj, FName);

    if FPropInfo = nil then
    begin
      p := Pos('.', FName);

      if p > 0 then
      begin
        // Property is a sub property of an object such as Font.Charset
        objNames := TStringList.Create;
        try
          propName := FName;

          while p > 0 do
          begin
            objNames.Add(Copy(propName, 1, p - 1));
            Delete(propName, 1, p);
            p := Pos('.', propName);
          end;

          for i := 0 to objNames.Count - 1 do
          begin
            FObj := TObject(GetOrdProp(FObj, objNames[i]));

            if FObj = nil then
              Break;
          end;

          if FObj <> nil then
          begin
            FSubName := propName;
            FPropInfo := GetPropInfo(FObj, propName);
          end;
        finally
          objNames.Free;
        end;
      end
      else
      begin
        actualName := LaModules.GetActualName(FObj, FName);

        if (actualName <> '') and (actualName <> FName) then
          FPropInfo := GetPropInfo(FObj, actualName);
      end;
    end;
  end
  else
    FPropInfo := nil;

  ProcessPropertyValue;
end;

function TLaTranslator.GetPropValue(instance: TObject): Variant;
{$IFDEF VCL9}
var
  dynArray: Pointer;
{$ENDIF}
begin
  Result := Null;

  case FPropInfo^.PropType^^.Kind of
    tkInteger, tkClass:
      Result := GetOrdProp(instance, FPropInfo);

    tkChar:
      Result := Char(GetOrdProp(instance, FPropInfo));

    tkWChar:
      Result := WideChar(GetOrdProp(instance, FPropInfo));

    tkEnumeration:
      if GetTypeData(FPropInfo^.PropType^)^.BaseType^ = System.TypeInfo(Boolean) then
        Result := Boolean(GetOrdProp(instance, FPropInfo))
      else
        Result := GetOrdProp(instance, FPropInfo);

    tkSet:
      Result := GetOrdProp(Instance, FPropInfo);

    tkFloat:
      Result := GetFloatProp(Instance, FPropInfo);

    tkMethod:
      Result := FPropInfo^.PropType^.Name;

    tkString, tkLString:
      Result := GetStrProp(instance, FPropInfo);

{$IFDEF VCL6}
    tkWString:
      Result := GetWideStrProp(instance, FPropInfo);
{$ENDIF}

{$IFDEF UNICODE_VCL}
    tkUString:
      Result := GetUnicodeStrProp(instance, FPropInfo);
{$ENDIF}

    tkVariant:
      Result := GetVariantProp(instance, FPropInfo);

{$IFDEF VCL6}
    tkInt64:
      Result := GetInt64Prop(instance, FPropInfo);
{$ENDIF}

    tkDynArray:
      begin
{$IFDEF VCL9}
        DynArray := GetDynArrayProp(instance, FPropInfo);
        DynArrayToVariant(Result, dynArray, FPropInfo^.PropType^);
{$ELSE}
        DynArrayToVariant(Result, Pointer(GetOrdProp(instance, FPropInfo)), FPropInfo^.PropType^);
{$ENDIF}
      end;
  end;
end;

function IsString(varType: TVarType): Boolean;
begin
  Result :=
    (varType = varString) or
{$IFDEF UNICODE_VCL}
    (varType = varUString) or
{$ENDIF}
    (varType = varOleStr);
  end;

procedure TLaTranslator.SetPropValue(
  instance: TObject;
  const value: Variant);

  procedure RangeError;
  begin
    raise ERangeError.CreateRes(@SRangeError);
  end;

  function RangedValue(const aMin, aMax: Int64): Int64;
  begin
    Result := Trunc(value);

    if (Result < aMin) or (Result > aMax) then
      RangeError;
  end;

  function RangedCharValue(const aMin, aMax: Int64): Int64;
  var
    ansi: AnsiString;
    wide: UnicodeString;
  begin
    case VarType(value) of
      varString:
        begin
          ansi := AnsiString(value);

          if Length(ansi) = 1 then
            Result := Ord(ansi[1])
          else
            Result := aMin - 1;
       end;

{$IFDEF UNICODE_VCL}
      varUString,
{$ENDIF}
      varOleStr:
        begin
          wide := value;

          if Length(wide) = 1 then
            Result := Integer(wide[1])
          else
            Result := aMin - 1;
        end;
    else
      Result := Trunc(value);
    end;

    if (Result < aMin) or (Result > aMax) then
      RangeError;
  end;

var
  typeData: PTypeData;
  dynArray: Pointer;
{$IFNDEF VCL10}
  oldFarEast: Boolean;
{$ENDIF}
begin
  typeData := GetTypeData(FPropInfo^.PropType^);

{$IFNDEF VCL10}
  oldFarEast := SysLocale.FarEast;

  if instance is TAction then
    SysLocale.FarEast := True;
{$ENDIF}

  case FPropInfo.PropType^^.Kind of
    tkChar, tkWChar:
      SetOrdProp(
        instance,
        FPropInfo,
        RangedCharValue(typeData^.MinValue, typeData^.MaxValue));

    tkInteger:
      if typeData^.MinValue < typeData^.MaxValue then
        SetOrdProp(
          instance,
          FPropInfo,
          RangedValue(typeData^.MinValue, typeData^.MaxValue))
      else
        SetOrdProp(
          instance,
          FPropInfo,
          RangedValue(LongWord(typeData^.MinValue), LongWord(typeData^.MaxValue)));

    tkEnumeration:
      if IsString(value) then
        SetEnumProp(instance, FPropInfo, VarToStr(value))
      else if VarType(value) = varBoolean then
        SetOrdProp(instance, FPropInfo, Abs(Trunc(value)))
      else
        SetOrdProp(instance, FPropInfo, RangedValue(typeData^.MinValue, typeData^.MaxValue));

    tkSet:
      if VarType(value) = varInteger then
        SetOrdProp(instance, FPropInfo, value)
      else
        SetSetProp(instance, FPropInfo, VarToStr(value));

    tkFloat:
      SetFloatProp(instance, FPropInfo, value);

    tkString,
    tkLString:
      SetStrProp(instance, FPropInfo, VarToStr(value));

{$IFDEF VCL6}
    tkWString:
      SetWideStrProp(instance, FPropInfo, VarToWideStr(value));
{$ENDIF}

{$IFDEF UNICODE_VCL}
    tkUString:
      SetUnicodeStrProp(instance, FPropInfo, VarToWideStr(value));
{$ENDIF}

    tkVariant:
      SetVariantProp(instance, FPropInfo, value);

    tkInt64:
      SetInt64Prop(
        instance,
        FPropInfo,
        RangedValue(typeData^.MinInt64Value, typeData^.MaxInt64Value));

    tkDynArray:
    begin
      dynArray := nil;
      DynArrayFromVariant(dynArray, value, FPropInfo^.PropType^);
      SetOrdProp(instance, FPropInfo, Integer(dynArray));
    end;
  end;

{$IFNDEF VCL10}
  if instance is TAction then
    SysLocale.FarEast := oldFarEast;
{$ENDIF}
end;

procedure TLaTranslator.ProcessPropertyValue;
var
  current, value: Variant;

  function TypesEqual: Boolean;
  var
    currentType, newType: TVarType;
  begin
    currentType := VarType(current);
    newType := VarType(value);

    Result :=
      (currentType = newType) or
      (IsString(currentType) and IsString(newType));
  end;

  function IgnoreProperty: Boolean;
  var
    typeKind: TypInfo.TTypeKind;
  begin
    // Charset property is always translated.
    // If there are no memebers in LaEnabledProperties it means to translate
    // all properties.
    if (FName = 'Charset') or (LaEnabledProperties = []) then
      Result := False
    else
    begin
      if FPropInfo <> nil then
        typeKind := FPropInfo.PropType^.Kind
      else
        typeKind := tkLString;

      if FTranslateLayout and (typeKind in [tkInteger]) then
        Result := False
      else
        Result := not (typeKind in LaEnabledProperties);
    end;

    // By default Left, Top and Position properties of the form and
    // ConnectionString property of any other component is not translated.
    // If you want to translate postion of form turn
    // LaFormPositionTranslationEnabled to True.
    if not Result then
    begin
      if FObj = FHost then
        Result :=
          not LaFormPositionTranslationEnabled and
          ((FName = 'Position') or (FName = 'Left') or (FName = 'Top'))
      else
        Result := (FName = 'ConnectionString');
    end;

    if Assigned(LaBeforeTranslate) then
      LaBeforeTranslate(FHost, FObj, FPropInfo, current, value, Result);
  end;

var
  int: Longint;
  str: String;
  identToInt: TIdentToInt;
  module: TLaModule;
begin
  case FReader.NextValue of
    vaNull:
      FReader.ReadValue;

    vaList:
      ProcessList;

    vaCollection:
      ProcessCollection;

    vaBinary:
      value := ProcessBinary;

    vaSet:
      ProcessSet;

    vaIdent:
    begin
      str := FReader.ReadIdent;

      if TypeInfo <> nil then
      begin
        case TypeInfo.Kind of
          tkEnumeration:
            value := GetEnumValue(TypeInfo, str);

          tkInteger:
          begin
           identToInt := FindIdentToInt(TypeInfo);

           if Assigned(identToInt) and IdentToInt(str, int) then
             value := int;
          end;
        end;
      end;
    end;

    vaInt8,
    vaInt16,
    vaInt32:
      value := FReader.ReadInteger;

    vaExtended,
    vaSingle:
      value := FReader.ReadFloat;

    vaString,
    vaLString:
      value := FReader.ReadString;

    vaWString:
      value := FReader.ReadWideString;

    vaFalse,
    vaTrue:
      value := FReader.ReadBoolean;

    vaNil:
      ProcessBytes(0);

    vaCurrency:
      value := FReader.ReadCurrency;

    vaDate:
      ProcessBytes(Sizeof(TDateTime));

    vaInt64:
      ProcessBytes(Sizeof(Int64));

{$IFDEF VCL6}
    vaUTF8String:
      value := FReader.ReadWideString;
{$ENDIF}

{$IFDEF VCL9}
    vaDouble:
      value := FReader.ReadDouble;
{$ENDIF}
  else
    raise Exception.Create('Unknown value type');
  end;

  if not VarIsNull(value) and not VarIsEmpty(value) then
  begin
    if FPropInfo <> nil then
    begin
      // Property is a simple property. Use generic translation method.
      // Get the current value, if it is different to the new value set the property
      try
        current := GetPropValue(FObj);

        if TypesEqual and (current <> value) and not IgnoreProperty then
        begin
          SetPropValue(FObj, value);

          if Assigned(LaAfterTranslate) then
            LaAfterTranslate(FHost, FObj, FPropInfo);
        end;
      except
        on e: Exception do
          raise Exception.CreateFmt('Could not translate %s.%s: %s', [FObj.ClassName, FName, e.Message]);
      end;
    end
    else if LaModules.CanTranslate(FObj, module) then
      module.Translate(FCurrent, FObj, FPropertyName, value, FIndex);
  end;
end;

procedure TLaTranslator.ProcessComponent(parent: TComponent; root: Boolean);

  function FindUnnamedComponent(
    parent: TComponent;
    const typeName: String;
    index: Integer): TComponent;
  var
    i, thisIndex: Integer;
    component: TComponent;
  begin
    Result := nil;
    thisIndex := 0;

    for i := 0 to parent.ComponentCount - 1 do
    begin
      component := parent.Components[i];

      if (component.Name = '') and (component.ClassName = typeName) then
      begin
        if thisIndex = index then
        begin
          Result := component;
          Break;
        end;

        Inc(thisIndex);
      end;
    end;
  end;

var
  i, childPos, index: Integer;
  typeName, componentName: String;
  flags: TFilerFlags;
  thisComponent: TComponent;
begin
  FReader.ReadPrefix(flags, childPos);
  typeName := FReader.ReadStr;
  componentName := FReader.ReadStr;

  if root then
    thisComponent := FHost
  else if parent = nil then
    thisComponent := nil
  else if componentName = '' then
  begin
    index := 0;

    for i := 0 to FUnnamedTypes.Count - 1 do
      if FUnnamedTypes[i] = typeName then
        Inc(index);

    FUnnamedTypes.Add(typeName);

    thisComponent := FindUnnamedComponent(parent, typeName, index);

    if thisComponent = nil then
      thisComponent := FindUnnamedComponent(FHost, typeName, index);
  end
  else
  begin
    thisComponent := parent.FindComponent(componentName);

    if thisComponent = nil then
      thisComponent := FHost.FindComponent(componentName);
  end;

  FCurrent := thisComponent;

  while not FReader.EndOfList do
    ProcessProperty(thisComponent);

  FReader.ReadListEnd;

  while not FReader.EndOfList do
    ProcessComponent(thisComponent, False);

  FReader.ReadListEnd;

  if thisComponent is TControl then
    TControl(thisComponent).Perform(CM_TEXTCHANGED, 0, 0);
end;

function FindInstance(classType: TClass): LongWord;
begin
  Result := FindResourceHInstance(FindClassHInstance(classType));
end;

procedure TLaTranslator.DoTranslate(component: TComponent; resourceName: String);
var
  instance: LongWord;
  header: array[0..3] of AnsiChar;
  stream: TResourceStream;
begin
  FHost := component;

  if resourceName = '' then
    resourceName := component.ClassName;

  instance := FindInstance(component.ClassType);

  if FindResource(instance, PChar(resourceName), RT_RCDATA) <> 0 then
  begin
    stream := TResourceStream.Create(instance, PChar(resourceName), RT_RCDATA);
    FReader := TReader.Create(stream, stream.Size);
    try
      FReader.Read(header, 4);

      if header = VCL_FORM_HEADER_C then
        ProcessComponent(nil, True);
    finally
      FReader.Free;
      stream.Free;
    end;
  end;
end;

procedure TLaTranslator.Translate(component: TComponent);

  function DoesResourceExists(classType: TClass): Boolean;
  var
    resourceName: String;
  begin
    resourceName := classType.ClassName;
    Result := FindResource(FindInstance(classType), PChar(resourceName), RT_RCDATA) <> 0;
  end;

var
  i: Integer;
  thisClass: TClass;
  names: TStringList;
begin
  names := TStringList.Create;
  try
    thisClass := component.ClassType;

    while DoesResourceExists(thisClass) do
    begin
      names.Insert(0, thisClass.ClassName);
      thisClass := thisClass.ClassParent;
    end;

    for i := 0 to names.Count - 1 do
      DoTranslate(component, names[i]);
  finally
    names.Free;
  end;
end;

procedure TLaTranslator.TranslateForm(form: TForm);
var
  i: Integer;
begin
  if (roFlipChildren in LaResourceOptions) and IsPreviousLocaleBidi then
    form.FlipChildren(True);

  for i := 0 to form.ComponentCount - 1 do
    if form.Components[i] is TFrame then
      Translate(form.Components[i]);

  Translate(form);

  if (roFlipChildren in LaResourceOptions) and IsActiveLocaleBidi then
    form.FlipChildren(True);
end;

class procedure TLaTranslator.TranslateForms;
var
  translator: TLaTranslator;

  procedure SetNewLayout(value: TLaUiLayout);
  begin
    translator.FTranslateLayout := UiLayout <> value;
    UiLayout := value;
  end;

var
  i: Integer;
begin
  translator := TLaTranslator.Create;
  try
    // Checks if layout will change. If it does force to translate layout.
    if IsActiveLocaleBidi and (UiLayout = ulLeftToRight) then
      SetNewLayout(ulRightToLeft)
    else if not IsActiveLocaleBidi and (UiLayout = ulRightToLeft) then
      SetNewLayout(ulLeftToRight);

    // Translates data modules
    for i := 0 to Screen.DataModuleCount - 1 do
      translator.Translate(Screen.DataModules[i]);

    // Translates forms
    for i := 0 to Screen.FormCount - 1 do
      translator.TranslateForm(Screen.Forms[i]);
  finally
    translator.Free;
  end;
end;


// TLaStream

constructor TLaStream.Create(const value: RawByteString);
begin
  inherited Create;

  if Length(value) > 0 then
  begin
    Write(PAnsiChar(value)^, Length(value));
    Seek(0, soFromBeginning);
  end;
end;

function TLaStream.ReadByte: Byte;
begin
  Read(Result, SizeOf(Result));
end;

function TLaStream.ReadWord: Word;
begin
  Read(Result, SizeOf(Result));
end;

function TLaStream.ReadInteger: Integer;
begin
  Read(Result, SizeOf(Result));
end;

function TLaStream.ReadPointer: Pointer;
begin
  Read(Result, SizeOf(Result));
end;

function TLaStream.ReadShortString: RawByteString;
var
  size: Byte;
begin
  Read(size, SizeOf(size));
  SetLength(Result, size);
  Read(PAnsiChar(Result)^, size);
end;

function TLaStream.ReadShortUnicodeString: UnicodeString;
var
  size: Byte;
begin
  Read(size, SizeOf(size));
  SetLength(Result, size);
  Read(PWideChar(Result)^, 2*size);
end;


// TLaModule

function GetHost(component: TComponent): TComponent;
begin
  if (component.Owner <> nil) and ((component.Owner is TCustomForm) or (component.Owner is TDataModule)) then
    Result := component.Owner
  else
    Result := component;
end;

function TLaModule.IgnoreProperty(
  component: TComponent;
  obj: TObject;
  current: Variant;
  var value: Variant;
  propInfo: PPropInfo): Boolean;
begin
  Result := False;

  if Assigned(LaBeforeTranslate) then
    LaBeforeTranslate(GetHost(component), obj, propInfo, current, value, Result);
end;

procedure TLaModule.AfterTranslate(
  component: TComponent;
  obj: TObject;
  propInfo: PPropInfo);
begin
  if Assigned(LaAfterTranslate) then
    LaAfterTranslate(GetHost(component), obj, nil);
end;

function TLaModule.GetActualObject(
  obj: TObject;
  const propName: String): TObject;
begin
  Result := nil;
end;

function TLaModule.GetActualName(
  obj: TObject;
  const propName: String): String;
begin
  Result := '';
end;


// TLaStringsModule

function TLaStringsModule.CanTranslate(obj: TObject): Boolean;
begin
  Result := obj is TStrings;
end;

procedure TLaStringsModule.Translate(
  component: TComponent;
  obj: TObject;
  const name: String;
  value: Variant;
  index: Integer);
var
  oldItemIndex: Integer;
  strings: TStrings;
begin
  strings := obj as TStrings;

  if (index >= strings.Count) or IgnoreProperty(component, obj, strings[index], value) then
    Exit;

{$IFDEF VCL6}
  if component is TCustomListControl then
    oldItemIndex := TCustomListControl(component).ItemIndex
{$ELSE}
  if component is TCustomComboBox then
    oldItemIndex := TCustomComboBox(component).ItemIndex
{$ENDIF}
  else
    oldItemIndex := -1;

  strings[index] := value;

{$IFDEF VCL6}
  if component is TCustomListControl then
    TCustomListControl(component).ItemIndex := oldItemIndex;
{$ELSE}
  if component is TCustomComboBox then
    TCustomComboBox(component).ItemIndex := oldItemIndex;
{$ENDIF}

  AfterTranslate(component, obj);
end;


// TLaModules

constructor TLaModules.Create;
begin
  inherited;
  FItems := TList.Create;
end;

destructor TLaModules.Destroy;
begin
  ClearItems;
  FItems.Free;
  inherited;
end;

function TLaModules.GetCount: Integer;
begin
  Result := FItems.Count;
end;

function TLaModules.GetItem(i: Integer): TLaModule;
begin
  Result := FItems[i];
end;

procedure TLaModules.Register(moduleClass: TLaModuleClass);
var
  i: Integer;
begin
  for i := 0 to Count - 1 do
    if Items[i].ClassType = moduleClass then
      Exit;

  FItems.Add(moduleClass.Create);
end;

function TLaModules.CanTranslate(obj: TObject; var module: TLaModule): Boolean;
var
  i: Integer;
begin
  for i := 0 to FItems.Count - 1 do
  begin
    module := Items[i];
    Result := module.CanTranslate(obj);

    if Result then
      Exit;
  end;

  module := nil;
  Result := False;
end;

function TLaModules.GetActualObject(
  obj: TObject;
  const propName: String): TObject;
var
  i: Integer;
begin
  for i := 0 to FItems.Count - 1 do
  begin
    Result := Items[i].GetActualObject(obj, propName);

    if Result <> nil then
      Exit;
  end;

  Result := nil;
end;

function TLaModules.GetActualName(obj: TObject; const propName: String): String;
var
  i: Integer;
begin
  for i := 0 to FItems.Count - 1 do
  begin
    Result := Items[i].GetActualName(obj, propName);

    if Result <> '' then
      Exit;
  end;

  Result := '';
end;

procedure TLaModules.ClearItems;
begin
  while FItems.Count > 0 do
  begin
    TObject(FItems[0]).Free;
    FItems.Delete(0);
  end;
end;

initialization
  LaEnabledProperties := [];
  LaFormPositionTranslationEnabled := False;
  LaBeforeTranslate := nil;
  LaAfterTranslate := nil;

  LaModules := TLaModules.Create;
  LaModules.Register(TLaStringsModule);
finalization
  LaModules.Free;
end.
